'use strict';

var t = require('@babel/types');
var core = require('@taddy/core');
var taddy = require('taddy');
var evaluate = require('./helpers/evaluate.cjs');
var getObjectPropertyKey = require('./helpers/getObjectPropertyKey.cjs');
var optimizeStaticStyles = require('./helpers/optimizeStaticStyles.cjs');
var mergeObjectProperties = require('./helpers/mergeObjectProperties.cjs');
var TSProcessor = require('./TSProcessor.cjs');

function _interopNamespaceDefault(e) {
    var n = Object.create(null);
    if (e) {
        Object.keys(e).forEach(function (k) {
            if (k !== 'default') {
                var d = Object.getOwnPropertyDescriptor(e, k);
                Object.defineProperty(n, k, d.get ? d : {
                    enumerable: true,
                    get: function () { return e[k]; }
                });
            }
        });
    }
    n.default = e;
    return Object.freeze(n);
}

var t__namespace = /*#__PURE__*/_interopNamespaceDefault(t);

function isUndefined(path) {
    return path.isIdentifier() && path.node.name === 'undefined';
}
function getLiteralValue(path) {
    try {
        return new Function(`return ${path.toString()}`)();
    }
    catch (error) {
        return getLiteralValue.FAIL;
    }
}
getLiteralValue.FAIL = Symbol('FAILED_VALUE');
function getHashedName(key, { postfix }) {
    return core.config.nameGenerator.getName(key, '', { postfix }).join('');
}
const isCSSProperty = (key) => {
    return key.slice(0, 2) === '--' || /[\w-]/.test(key);
};
// TODO: bottom up and split the initialization
let cachedTSProject;
class Processor {
    config;
    options;
    isStatic;
    classNameNodes;
    optimizationPaths;
    variables;
    prepare(options) {
        this.options = options;
        this.variables = [];
        this.classNameNodes = new WeakSet();
        this.optimizationPaths = new Set();
    }
    constructor({ config }) {
        this.config = config;
    }
    get tsProject() {
        if (!this.config.typescript) {
            throw new Error('TADDY: cant use typescript without enabled option');
        }
        return (cachedTSProject ||
            (cachedTSProject = new TSProcessor.TSProcessor(this.config.typescript === true ? {} : this.config.typescript)));
    }
    $css(rule, options) {
        // console.log('sm', this.options.sourceMap)
        const result = taddy.$css(rule, {
            ...options,
            // hash: stringHash(this.options.filename),
        });
        // console.trace({rule, options, result});
        return result;
    }
    evaluate(currentPath) {
        return evaluate.evaluate(currentPath, { exec: this.config.evaluate });
    }
    isClassNameNode(node) {
        return this.classNameNodes.has(node);
    }
    classNamesToNode(classNames) {
        const properties = [];
        for (const key in classNames) {
            const classValue = classNames[key];
            const propName = t__namespace.stringLiteral(key);
            let propValue;
            if (typeof classNames[key] === 'object') {
                propValue = t__namespace.objectExpression(this.classNamesToNode(classNames[key]));
            }
            else if (typeof classNames[key] === 'boolean') {
                propValue = t__namespace.booleanLiteral(classValue);
            }
            else {
                propValue = t__namespace.stringLiteral(String(classValue));
            }
            const node = t__namespace.objectProperty(propName, propValue);
            this.classNameNodes.add(node);
            properties.push(node);
        }
        return properties;
    }
    stylesToNode(key, value, { postfix = '', at } = {}) {
        const { className } = this.$css({ [key]: value }, { postfix, at });
        const properties = this.classNamesToNode(className);
        return properties;
    }
    processAtRule(key, path, { postfix, properties }) {
        const atRuleName = key;
        const valuePath = path.get('value');
        // if it's not an object expression, we can treat it as a regular object value
        if (!valuePath.isObjectExpression()) {
            return this.processPropertyValue(key, path, { postfix, properties });
        }
        for (const propPath of valuePath.get('properties')) {
            if (propPath.isObjectProperty()) {
                this.processObjectProperty(propPath, {
                    postfix,
                    at: { name: atRuleName },
                    properties,
                });
                continue;
            }
            if (propPath.isSpreadElement()) {
                this.processSpreadElement(propPath, {
                    postfix,
                    at: { name: atRuleName },
                    properties,
                });
                continue;
            }
            properties.push(propPath.node);
        }
    }
    processPropertyValue(key, path, { postfix, at, properties }) {
        const keyPath = path.get('key');
        const valuePath = path.get('value');
        const tryLiteralValue = () => {
            if (!(valuePath.isLiteral() || isUndefined(valuePath)))
                return false;
            const value = getLiteralValue(valuePath);
            if (value === getLiteralValue.FAIL) {
                return false;
            }
            properties.push(...this.stylesToNode(key, value, { postfix, at }));
            return true;
        };
        const tryObjectExpressionValue = () => {
            if (!valuePath.isObjectExpression())
                return false;
            // we shouldn't share "properties", as it should construct a new style object
            this.process(valuePath, {
                postfix: postfix + key,
                at,
            });
            // TODO: add comments
            let currentStyles = [];
            for (const nestedProp of valuePath.node.properties) {
                if (this.isClassNameNode(nestedProp)) {
                    if (currentStyles.length) {
                        properties.push(t__namespace.objectProperty(keyPath.node, t__namespace.objectExpression(currentStyles)));
                        currentStyles = [];
                    }
                    properties.push(nestedProp);
                }
                else {
                    currentStyles.push(nestedProp);
                }
            }
            if (currentStyles.length) {
                properties.push(t__namespace.objectProperty(keyPath.node, t__namespace.objectExpression(currentStyles)));
            }
            return true;
        };
        const tryComposesValue = () => {
            if (!(key === 'composes' && valuePath.isArrayExpression())) {
                return false;
            }
            if (postfix !== '') {
                throw new Error('TADDY: composes could be used only on top level');
            }
            for (const elemPath of valuePath.get('elements')) {
                this.process(elemPath, { at, postfix, properties });
            }
            return false;
        };
        const tryEvaluateValue = () => {
            const result = this.evaluate(valuePath);
            if ('error' in result)
                return false;
            properties.push(...this.stylesToNode(key, result.value, { postfix, at }));
            this.optimizationPaths.add(path);
            return true;
        };
        const STOP = Symbol('STOP');
        const getTSValue = (tsValuePath) => {
            const tsType = this.tsProject?.getTypeAtPos(this.options.filename, this.options.code, tsValuePath.node.start || 0, tsValuePath.node.end || 0);
            if (!tsType)
                return STOP;
            return TSProcessor.parseValue(tsType.getType());
        };
        const getComposesTSValue = (path) => {
            const elements = [];
            elements[TSProcessor.STATIC_KEY] = true;
            for (const elementPath of path.get('elements')) {
                if (!elementPath)
                    return;
                const value = getTSValue(elementPath);
                if (value === STOP)
                    return;
                elements[TSProcessor.STATIC_KEY] =
                    elements[TSProcessor.STATIC_KEY] && TSProcessor.isStaticValue(value);
                elements.push(value);
            }
            return elements;
        };
        const tryTypescriptValue = () => {
            if (!this.config.typescript)
                return false;
            let tsValue;
            if (key === 'composes' && valuePath.isArrayExpression()) {
                tsValue = getComposesTSValue(valuePath);
            }
            else {
                tsValue = getTSValue(valuePath);
            }
            if (tsValue === STOP)
                return false;
            if (TSProcessor.isStaticValue(tsValue)) {
                properties.push(...this.stylesToNode(key, tsValue, { postfix, at }));
                this.optimizationPaths.add(path);
                return true;
            }
            // TODO: improve composes handling
            if (key === 'composes')
                return false;
            const variants = [].concat(tsValue);
            let isCompilable = true;
            // pre-generate classes for variants
            variants.forEach((v) => {
                if (v instanceof TSProcessor.DynamicType || !TSProcessor.isStaticValue(v)) {
                    isCompilable = false;
                    return;
                }
                this.$css({ [key]: v }, { postfix, at });
            });
            if (!isCompilable)
                return false;
            const propKey = getHashedName(key, { postfix, at });
            const hashFunctionNode = this.options.addImport('h');
            properties.push(t__namespace.objectProperty(t__namespace.identifier(propKey), t__namespace.callExpression(hashFunctionNode, [
                valuePath.node,
            ])));
            return true;
        };
        const tryCSSVariableValue = () => {
            if (!this.config.CSSVariableFallback)
                return false;
            // think about dynamic nested for mixins
            // in case of nested dynamic values in mixin we can't fallback them as custom properties
            if (this.options.mixin ||
                key === 'composes' ||
                !isCSSProperty(key)) {
                return false;
            }
            const value = getHashedName(key, { postfix, at });
            const dynamicValue = `--${value}`;
            this.variables.push(t__namespace.objectProperty(t__namespace.stringLiteral(dynamicValue), valuePath.node));
            properties.push(...this.stylesToNode(key, `var(${dynamicValue})`, {
                postfix,
                at,
            }));
            return true;
        };
        if (tryLiteralValue() ||
            tryObjectExpressionValue() ||
            tryComposesValue() ||
            tryEvaluateValue() ||
            tryTypescriptValue() ||
            tryCSSVariableValue()) {
            return;
        }
        properties.push(path.node);
    }
    processObjectProperty(path, { postfix = '', at, properties }) {
        const keyPath = path.get('key');
        const valuePath = path.get('value');
        const key = getObjectPropertyKey.getObjectPropertyKey(keyPath);
        const tryEvaluateKey = () => {
            const result = this.evaluate(keyPath);
            if ('error' in result)
                return false;
            keyPath.replaceWith(t__namespace.stringLiteral(result.value));
            this.processPropertyValue(result.value, path, {
                postfix,
                at,
                properties,
            });
            return true;
        };
        if (key === 'className' || key === 'style' || key === null) {
            if (tryEvaluateKey()) {
                return;
            }
            properties.push(path.node);
            return;
        }
        if (key[0] === '@') {
            this.processAtRule(key, path, { postfix, properties });
            return;
        }
        if (at?.name && !at?.query) {
            this.process(valuePath, {
                postfix,
                properties,
                at: { name: at.name, query: key },
            });
            return;
        }
        this.processPropertyValue(key, path, { postfix, at, properties });
    }
    processSpreadElement(path, { postfix = '', at, properties }) {
        const tryEvaluateSpreadElement = () => {
            const result = this.evaluate(path.get('argument'));
            if ('error' in result)
                return false;
            const { className } = this.$css(result.value, { postfix, at });
            properties.push(...this.classNamesToNode(className));
            this.optimizationPaths.add(path);
            return true;
        };
        const tryTypescriptSpreadElement = () => {
            if (!this.config.typescript)
                return false;
            const { argument } = path.node;
            const tsType = this.tsProject?.getTypeAtPos(this.options.filename, this.options.code, argument.start || 0, argument.end || 0);
            if (!tsType)
                return false;
            const styles = TSProcessor.parseValue(tsType.getType());
            if (!TSProcessor.isStaticValue(styles))
                return false;
            const { className } = this.$css(styles, { postfix, at });
            properties.push(...this.classNamesToNode(className));
            this.optimizationPaths.add(path);
            return true;
        };
        if (tryEvaluateSpreadElement() || tryTypescriptSpreadElement()) {
            return;
        }
        // we shouldn't share "properties", as it should construct a new style object
        this.process(path.get('argument'), { postfix, at });
        properties.push(path.node);
    }
    processLogicalExpression(path, options) {
        this.process(path.get('left'), options);
        this.process(path.get('right'), options);
    }
    processObjectExpression(path, { postfix = '', at, properties = [] }) {
        for (const propPath of path.get('properties')) {
            if (propPath.isObjectProperty()) {
                this.processObjectProperty(propPath, { postfix, at, properties });
                continue;
            }
            if (propPath.isSpreadElement()) {
                this.processSpreadElement(propPath, { postfix, at, properties });
                continue;
            }
            properties.push(propPath.node);
        }
        /**
         * TODO: think about static optimizations for mixins
         */
        if (this.options.mixin)
            return;
        path.node.properties = mergeObjectProperties.mergeObjectProperties(properties);
    }
    processConditionalExpression(path, options = {}) {
        this.process(path.get('consequent'), options);
        this.process(path.get('alternate'), options);
    }
    process(path, options = {}) {
        if (path.isLogicalExpression()) {
            this.processLogicalExpression(path, options);
        }
        else if (path.isObjectExpression()) {
            this.processObjectExpression(path, options);
        }
        else if (path.isConditionalExpression()) {
            this.processConditionalExpression(path, options);
        }
    }
    run(callPath, options) {
        this.prepare(options);
        const tryEvaluate = (path) => {
            const result = this.evaluate(path);
            if ('error' in result)
                return false;
            const { className } = this.$css(result.value);
            /**
             * TODO: think about static optimizations for mixins
             */
            if (this.options.mixin)
                return true;
            const properties = this.classNamesToNode(className);
            path.replaceWith(t__namespace.objectExpression(mergeObjectProperties.mergeObjectProperties(properties)));
            this.optimizationPaths.add(path);
            return true;
        };
        const args = [].concat(callPath.get('arguments'));
        let isObjectable = true;
        for (const arg of args) {
            if (tryEvaluate(arg)) {
                continue;
            }
            this.process(arg);
            if (!arg.isObjectExpression()) {
                isObjectable = false;
                continue;
            }
        }
        const nodes = args.map((x) => x.node);
        const argPath = args[0];
        if (this.variables.length > 0) {
            nodes.push(t__namespace.objectExpression([
                t__namespace.objectProperty(t__namespace.identifier(core.VARS_KEY), t__namespace.objectExpression(this.variables)),
            ]));
        }
        if (isObjectable) {
            argPath.replaceWith(t__namespace.objectExpression(mergeObjectProperties.mergeObjects(nodes)));
            callPath.node.arguments = [argPath.node];
            const path = argPath;
            const isStatic = path.node.properties.every((x) => {
                if (!t__namespace.isObjectProperty(x))
                    return false;
                if (!('value' in x.key))
                    return false;
                if (!(x.key.value[0] === '_' ||
                    x.key.value === 'className' ||
                    x.key.value === 'style' ||
                    x.key.value === core.VARS_KEY ||
                    t__namespace.isBooleanLiteral(x.value))) {
                    return false;
                }
                return true;
            });
            if (!this.options.mixin && isStatic) {
                optimizeStaticStyles.optimizeStaticStyles(path);
            }
            return {
                isStatic,
                optimizationPaths: this.optimizationPaths,
            };
        }
        return {
            isStatic: false,
            optimizationPaths: this.optimizationPaths,
        };
    }
}

exports.Processor = Processor;
//# sourceMappingURL=Processor.cjs.map
