import * as t from '@babel/types';
import assert from 'node:assert';
import path__default from 'node:path';
import { $css } from 'taddy';
import { getEnv, PACKAGE_NAME, MACRO_NAME } from './config.js';
import { isTaddyEvaluation } from './helpers/utils.js';
import { taggedTemplateToObject } from './helpers/taggedTemplateToObject.js';
import { createHandlers } from './handlers.js';
import { Output } from './Output.js';
import { makeSourceMapGenerator, convertGeneratorToComment } from './source-maps.js';

function mapCompileOptions({ filename, evaluate = true, unstable_typescript = false, unstable_CSSVariableFallback = true, unstable_optimizeBindings = true, unstable_useTaggedTemplateLiterals = false, unstable_target = 'auto', }) {
    const getTarget = () => {
        if (unstable_target !== 'auto')
            return unstable_target;
        const extname = path__default.extname(filename);
        if (extname === '.vue')
            return 'vue';
        return 'auto';
    };
    return {
        evaluate,
        typescript: unstable_typescript,
        CSSVariableFallback: unstable_CSSVariableFallback,
        optimizeBindings: unstable_optimizeBindings,
        useTaggedTemplateLiterals: unstable_useTaggedTemplateLiterals,
        target: getTarget(),
    };
}
const findImportPath = (program) => {
    let importPath = undefined;
    program.traverse({
        ImportDeclaration(p) {
            const { node } = p;
            if (!(node.source.value === PACKAGE_NAME ||
                node.source.value === MACRO_NAME)) {
                return;
            }
            importPath = p;
            importPath.node.source = t.stringLiteral('taddy');
            p.stop();
        },
    });
    return importPath;
};
let output;
let sourceMapGenerator;
function macro({ references, babel, state, config: _config = {}, }) {
    const { env = getEnv(babel), ...config } = _config;
    if (isTaddyEvaluation(state)) {
        return;
    }
    const program = state.file.path;
    const code = state.file.code;
    const { filename, sourceRoot, root, filenameRelative = path__default.relative(sourceRoot || root || process.cwd(), filename), } = state.file.opts;
    assert(filename, 'Filename is required');
    assert(filenameRelative, 'FilenameRelative is required');
    const importPath = findImportPath(program);
    if (!importPath)
        return;
    sourceMapGenerator = makeSourceMapGenerator(state.file);
    sourceMapGenerator.setSourceContent(filename, code);
    // TODO: implement "createCSS" and pass it
    $css.ruleInjector.reset();
    const importCache = new Map();
    const { useTaggedTemplateLiterals, ...compileOptions } = mapCompileOptions({
        filename,
        ...config.compileOptions,
    });
    switch (compileOptions.target) {
        case 'vue': {
            require('taddy/vue');
            break;
        }
    }
    const { handlers, finish } = createHandlers(compileOptions, {
        state,
        sourceMapGenerator,
        env,
        filename,
        filenameRelative,
        code,
        addImport(name) {
            if (!importCache.has(name)) {
                const specifier = t.importSpecifier(t.identifier(`_${name}_`), t.identifier(name));
                importCache.set(name, specifier);
                importPath.node.specifiers.push(specifier);
            }
            return importCache.get(name).local;
        },
    });
    for (const key in references) {
        if (!(key in handlers))
            continue;
        for (const ref of references[key]) {
            let currentPath = ref.parentPath;
            let currentKey = key;
            if (key === 'css' &&
                currentPath?.isMemberExpression() &&
                'name' in currentPath.node.property &&
                currentPath.node.property.name === 'mixin') {
                currentPath = currentPath.parentPath;
                currentKey = 'mixin';
            }
            if (useTaggedTemplateLiterals &&
                (currentKey === 'css' || currentKey === 'mixin') &&
                currentPath?.isTaggedTemplateExpression()) {
                const obj = taggedTemplateToObject(currentPath);
                currentPath.replaceWith(t.callExpression(currentPath.node.tag, [obj]));
            }
            handlers[currentKey](currentPath);
        }
    }
    const { isStatic } = finish();
    if (isStatic) {
        importPath.node.source = t.stringLiteral('@taddy/core');
    }
    output = output || new Output({ env, config: config.outputOptions });
    const sourceMap = convertGeneratorToComment(sourceMapGenerator);
    const result = output.save({
        sourceMap,
        filenameRelative,
        target: compileOptions.target,
    });
    importPath.insertAfter(t.importDeclaration([], t.stringLiteral(result.importName)));
    return {
        keepImports: true,
    };
}

export { macro };
//# sourceMappingURL=macro-plugin.js.map
